@@ -1,5 +1,5 @@
 /*
-	Copyright (c) 2013-2022 Cong Xu
+	Copyright (c) 2013-2023, 2025 Cong Xu
 	All rights reserved.
 
 	Redistribution and use in source and binary forms, with or without
@@ -37,7 +37,6 @@
 
 #include "animated_counter.h"
 #include "autosave.h"
-#include "hiscores.h"
 #include "loading_screens.h"
 #include "menu_utils.h"
 #include "password.h"
@@ -60,7 +59,8 @@ static void CampaignIntroOnExit(GameLoopData *data);
 static void CampaignIntroInput(GameLoopData *data);
 static GameLoopResult CampaignIntroUpdate(GameLoopData *data, LoopRunner *l);
 static void CampaignIntroDraw(GameLoopData *data);
-GameLoopData *ScreenCampaignIntro(CampaignSetting *c, const GameMode gameMode, const CampaignEntry *entry)
+GameLoopData *ScreenCampaignIntro(
+	CampaignSetting *c, const GameMode gameMode, const CampaignEntry *entry)
 {
 	ScreenCampaignIntroData *data;
 	CMALLOC(data, sizeof *data);
@@ -89,7 +89,8 @@ static void CampaignIntroOnEnter(GameLoopData *data)
 	else
 	{
 		MusicPlayFromChunk(
-			&gSoundDevice.music, MUSIC_MENU, &sData->c->CustomSongs[MUSIC_MENU]);
+			&gSoundDevice.music, MUSIC_MENU,
+			&sData->c->CustomSongs[MUSIC_MENU]);
 	}
 }
 static void CampaignIntroOnExit(GameLoopData *data)
@@ -117,14 +118,16 @@ static GameLoopResult CampaignIntroUpdate(GameLoopData *data, LoopRunner *l)
 	{
 		LoopRunnerPop(l);
 	}
-	else if (!IsIntroNeeded(gCampaign.Entry.Mode) ||
+	else if (
+		!IsIntroNeeded(gCampaign.Entry.Mode) ||
 		sData->waitResult == EVENT_WAIT_OK)
 	{
 		// Switch to num players selection
 		LoopRunnerPush(
-			l, ScreenLoading(
-				   "Loading campaign...", true,
-				   NumPlayersSelection(&gGraphicsDevice, &gEventHandlers), true));
+			l,
+			ScreenLoading(
+				"Loading campaign...", true,
+				NumPlayersSelection(&gGraphicsDevice, &gEventHandlers), true));
 	}
 	return UPDATE_RESULT_OK;
 }
@@ -140,7 +143,14 @@ static void CampaignIntroDraw(GameLoopData *data)
 	// Display title + author
 	char *buf;
 	CMALLOC(buf, strlen(sData->c->Title) + strlen(sData->c->Author) + 16);
-	sprintf(buf, "%s by %s", sData->c->Title, sData->c->Author);
+	if (strlen(sData->c->Author))
+	{
+		sprintf(buf, "%s by %s", sData->c->Title, sData->c->Author);
+	}
+	else
+	{
+		strcpy(buf, sData->c->Title);
+	}
 	FontOpts opts = FontOptsNew();
 	opts.HAlign = ALIGN_CENTER;
 	opts.Area = gGraphicsDevice.cachedConfig.Res;
@@ -468,14 +478,6 @@ static void MissionSummaryOnEnter(GameLoopData *data)
 {
 	MissionSummaryData *mData = data->Data;
 
-	if (mData->completed && CanLevelSelect(mData->c->Entry.Mode))
-	{
-		AutosaveAdd(
-			&gAutosave, &mData->c->Entry, mData->m->index,
-			mData->m->NextMission, &gPlayerDatas);
-		AutosaveSave(&gAutosave, GetConfigFilePath(AUTOSAVE_FILE));
-	}
-
 	// Calculate bonus scores
 	// Bonuses only apply if at least one player has lived
 	const int accessBonus = GetAccessBonus(mData->m);
@@ -498,6 +500,14 @@ static void MissionSummaryOnEnter(GameLoopData *data)
 		CA_FOREACH_END()
 	}
 
+	if (mData->completed && CanLevelSelect(mData->c->Entry.Mode))
+	{
+		AutosaveAdd(
+			&gAutosave, &mData->c->Entry, mData->m->index,
+			mData->m->NextMission, &gPlayerDatas);
+		AutosaveSave(&gAutosave, GetConfigFilePath(AUTOSAVE_FILE));
+	}
+
 	// Skip menu
 	if (mData->m->missionData->SkipDebrief)
 	{
@@ -630,11 +640,6 @@ static GameLoopResult MissionSummaryUpdate(GameLoopData *data, LoopRunner *l)
 			gCampaign.IsQuit = mData->ms.current->u.returnCode == 1;
 			LoopRunnerPop(l);
 		}
-		else
-		{
-			LoopRunnerChange(
-				l, HighScoresScreen(&gCampaign, &gGraphicsDevice));
-		}
 	}
 	return result;
 }
@@ -691,8 +696,9 @@ static void ApplyBonuses(PlayerData *p, const int bonus)
 }
 static int GetHealthBonus(const PlayerData *p)
 {
-	const int maxHealth = ModeMaxHealth(gCampaign.Entry.Mode);
-	return p->hp > maxHealth - 50 ? (p->hp + 50 - maxHealth) * 10 : 0;
+	const int maxHealth = p->Char.maxHealth;
+	const int hp = MIN(p->hp, maxHealth);
+	return hp > maxHealth - 50 ? (hp + 50 - p->Char.maxHealth) * 10 : 0;
 }
 static int GetResurrectionFee(const PlayerData *p)
 {
@@ -916,17 +922,27 @@ static void DrawObjectiveInfo(const Objective *o, const struct vec2i pos)
 		DrawHead(gGraphicsDevice.gameWindow.renderer, cd, DIRECTION_DOWN, pos);
 	}
 	break;
-	case OBJECTIVE_COLLECT:
+	case OBJECTIVE_COLLECT: {
+		struct vec2i drawPos = pos;
+		CA_FOREACH(const PickupClass *, pc, o->u.Pickups)
 		CPicDraw(
-			&gGraphicsDevice, &o->u.Pickup->Pic,
-			svec2i_subtract(pos, svec2i(-4, -4)), NULL);
-		break;
+			&gGraphicsDevice, &(*pc)->Pic,
+			svec2i_subtract(drawPos, svec2i(-4, -4)), NULL);
+		drawPos = svec2i_add(drawPos, svec2i(4, 2));
+		CA_FOREACH_END()
+	}
+	break;
 	case OBJECTIVE_DESTROY: {
+		struct vec2i drawPos = pos;
+		CA_FOREACH(const MapObject *, mo, o->u.MapObjects)
 		struct vec2i picOffset;
-		const Pic *p = MapObjectGetPic(o->u.MapObject, &picOffset);
+		const Pic *p = MapObjectGetPic(*mo, &picOffset);
 		PicRender(
-			p, gGraphicsDevice.gameWindow.renderer, svec2i_add(pos, picOffset),
-			colorWhite, 0, svec2_one(), SDL_FLIP_NONE, Rect2iZero());
+			p, gGraphicsDevice.gameWindow.renderer,
+			svec2i_add(drawPos, picOffset), colorWhite, 0, svec2_one(),
+			SDL_FLIP_NONE, Rect2iZero());
+		drawPos = svec2i_add(drawPos, svec2i(4, 2));
+		CA_FOREACH_END()
 	}
 	break;
 	case OBJECTIVE_INVESTIGATE:
