@@ -22,7 +22,7 @@
 	This file incorporates work covered by the following copyright and
 	permission notice:
 
-	Copyright (c) 2013-2022 Cong Xu
+	Copyright (c) 2013-2025 Cong Xu
 	All rights reserved.
 
 	Redistribution and use in source and binary forms, with or without
@@ -55,6 +55,7 @@
 #include <string.h>
 
 #include "actor_fire.h"
+#include "actor_pickup.h"
 #include "actor_placement.h"
 #include "ai.h"
 #include "ai_coop.h"
@@ -74,7 +75,6 @@
 #include "material.h"
 #include "mission.h"
 #include "pic_manager.h"
-#include "pickup.h"
 #include "sounds.h"
 #include "thing.h"
 #include "triggers.h"
@@ -142,6 +142,18 @@ void UpdateActorState(TActor *actor, int ticks)
 		}
 		actor->petrified = MAX(0, actor->petrified - ticks);
 		actor->confused = MAX(0, actor->confused - ticks);
+
+		// Reset accumulated damage if FPS_FRAMELIMIT passed since taking
+		// damage
+		actor->damageCooldownTicks += ticks;
+		if (actor->accumulatedDamage)
+		{
+			if (actor->damageCooldownTicks >= FPS_FRAMELIMIT)
+			{
+				actor->accumulatedDamage = 0;
+				actor->damageCooldownTicks = 0;
+			}
+		}
 	}
 
 	actor->slideLock = MAX(0, actor->slideLock - ticks);
@@ -169,7 +181,7 @@ void UpdateActorState(TActor *actor, int ticks)
 				e.u.AddPickup.Pos = Vec2ToNet(actor->Pos);
 				GameEventsEnqueue(&gGameEvents, e);
 			}
-			else
+			else if (!gCampaign.Setting.BuyAndSell)
 			{
 				// Add an ammo pickup of the actor's gun
 				if (gCampaign.Setting.Ammo)
@@ -331,12 +343,12 @@ static bool TryMoveActor(TActor *actor, struct vec2 pos)
 	CASSERT(
 		!svec2_is_nearly_equal(actor->Pos, pos, EPSILON_POS),
 		"trying to move to same position");
+	actor->CanPickupSpecial = false;
 
 	// Don't check collisions for pilots
 	if (actor->vehicleUID == -1)
 	{
 		actor->hasCollided = true;
-		actor->CanPickupSpecial = false;
 
 		const struct vec2 oldPos = actor->Pos;
 		pos = GetConstrainedPos(&gMap, actor->Pos, pos, actor->thing.size);
@@ -466,8 +478,10 @@ static bool CanMeleeTarget(
 	if (!w->Gun || WeaponClassCanShoot(w->Gun))
 		return false;
 	// Check for melee damage if we are the owner of the actor
-	if ((gCampaign.IsClient || a->PlayerUID < 0) &&
-		!ActorIsLocalPlayer(a->uid))
+	// Is client: must be local player
+	// Not client: AI or local player
+	if (!ActorIsLocalPlayer(a->uid) &&
+		(gCampaign.IsClient || a->PlayerUID >= 0))
 		return false;
 	// Check if any barrel can fire and melee
 	for (int barrel = 0; barrel < WeaponClassNumBarrels(w->Gun); barrel++)
@@ -701,8 +715,8 @@ static void CheckRescue(const TActor *a)
 	if (a->PlayerUID < 0)
 		return;
 
-		// Check an area slightly bigger than the actor's size for rescue
-		// objectives
+	// Check an area slightly bigger than the actor's size for rescue
+	// objectives
 #define RESCUE_CHECK_PAD 2
 	const CollisionParams params = {
 		THING_IMPASSABLE, CalcCollisionTeam(true, a),
@@ -727,15 +741,16 @@ static void CheckRescue(const TActor *a)
 	}
 }
 
-void ActorHeal(TActor *actor, int health)
+void ActorHeal(TActor *actor, const int amount, const bool exceedMax)
 {
-	actor->health += health;
-	actor->health = MIN(actor->health, ActorGetCharacter(actor)->maxHealth);
+	actor->health += amount;
+	const int maxHealth = ActorGetMaxHeal(actor, exceedMax);
+	actor->health = MIN(actor->health, maxHealth);
 	AddParticle ap;
 	memset(&ap, 0, sizeof ap);
 	ap.Pos = actor->Pos;
 	ap.Z = 10;
-	for (int i = 0; i < MAX(health / 20, 1); i++)
+	for (int i = 0; i < MAX(amount / 20, 1); i++)
 	{
 		ap.Vel = svec2(RAND_FLOAT(-0.2f, 0.2f), RAND_FLOAT(-0.2f, 0.2f));
 		EmitterStart(&actor->healEffect, &ap);
@@ -809,6 +824,12 @@ bool ActorUsesAmmo(const TActor *actor, const int ammoId)
 	return false;
 }
 
+static void ActorSetChatter(TActor *a, const char *text, const int count)
+{
+	strcpy(a->Chatter, text);
+	a->ChatterCounter = count;
+}
+
 void ActorReplaceGun(const NActorReplaceGun rg)
 {
 	TActor *a = ActorGetByUID(rg.UID);
@@ -826,13 +847,16 @@ void ActorReplaceGun(const NActorReplaceGun rg)
 	Weapon w = WeaponCreate(wc);
 	memcpy(&a->guns[rg.GunIdx], &w, sizeof w);
 	// Switch immediately to picked up gun
-	const PlayerData *p = PlayerDataGetByUID(a->PlayerUID);
-	if (wc->Type != GUNTYPE_GRENADE || !PlayerHasGrenadeButton(p))
+	if (ActorCanSwitchToWeapon(a, rg.GunIdx))
 	{
 		a->gunIndex = rg.GunIdx;
 	}
-
 	SoundPlayAt(&gSoundDevice, wc->SwitchSound, a->Pos);
+	// Show gun switch text as chatter, but only for PCs
+	if (a->PlayerUID >= 0)
+	{
+		ActorSetChatter(a, wc->name, CHATTER_SWITCH_GUN);
+	}
 }
 
 int ActorFindGun(const TActor *a, const WeaponClass *wc)
@@ -884,12 +908,6 @@ int ActorGetNumGrenades(const TActor *a)
 	return count;
 }
 
-static void ActorSetChatter(TActor *a, const char *text, const int count)
-{
-	strcpy(a->Chatter, text);
-	a->ChatterCounter = count;
-}
-
 // Set AI state and possibly say something based on the state
 void ActorSetAIState(TActor *actor, const AIState s)
 {
@@ -995,10 +1013,9 @@ static bool ActorTryChangeDirection(
 {
 	const bool willChangeDirecton =
 		!actor->petrified && CMD_HAS_DIRECTION(cmd) &&
-		(!(cmd & CMD_BUTTON2) ||
-		 ConfigGetEnum(&gConfig, "Game.SwitchMoveStyle") !=
-			 SWITCHMOVE_STRAFE) &&
-		(!(prevCmd & CMD_BUTTON1) ||
+		(!Button2(cmd) || ConfigGetEnum(&gConfig, "Game.SwitchMoveStyle") !=
+							  SWITCHMOVE_STRAFE) &&
+		(!Button1(prevCmd) ||
 		 ConfigGetEnum(&gConfig, "Game.FireMoveStyle") != FIREMOVE_STRAFE);
 	const direction_e dir = CmdToDirection(cmd);
 	if (willChangeDirecton && dir != actor->direction)
@@ -1015,7 +1032,7 @@ static bool ActorTryChangeDirection(
 
 static bool ActorTryShoot(TActor *actor, const int cmd)
 {
-	const bool willShoot = !actor->petrified && (cmd & CMD_BUTTON1);
+	const bool willShoot = !actor->petrified && Button1(cmd);
 	if (willShoot)
 	{
 		FireWeapon(actor, ACTOR_GET_GUN(actor));
@@ -1050,97 +1067,163 @@ static bool TryGrenade(TActor *a, const int cmd)
 }
 
 static bool ActorTryMove(TActor *actor, int cmd, int ticks);
-void CommandActor(TActor *actor, int cmd, int ticks)
+int CommandActor(TActor *actor, int cmd, int ticks)
 {
-	actor->hasShot = false;
-	// If this is a pilot, command the vehicle instead
-	if (actor->vehicleUID != -1)
+	GameEvent e;
+	// If the actor is currently using a menu, control the menu instead
+	if (actor->pickupMenu.pickup)
 	{
-		TActor *vehicle = ActorGetByUID(actor->vehicleUID);
-		CommandActor(vehicle, cmd, ticks);
-	}
-	else if (actor->pilotUID == -1)
-	{
-		// If this is a vehicle and there's no pilot, do nothing
+		if (Button1(cmd) && !Button1(actor->lastCmd))
+		{
+			// Apply effects of pickup menu item and reset
+			if (actor->pickupMenu.index <
+				(int)actor->pickupMenu.effect->u.Menu.Items.size)
+			{
+				const PickupMenuItem *m = CArrayGet(
+					&actor->pickupMenu.effect->u.Menu.Items,
+					actor->pickupMenu.index);
+				bool canPickup = false;
+				const char *sound = "menu_enter";
+				CA_FOREACH(const PickupEffect, pe, m->Effects)
+				CASSERT(
+					pe->Type != PICKUP_MENU, "can't have nested menu effects");
+				canPickup = PickupApplyEffect(
+					actor, actor->pickupMenu.pickup, pe, true, &sound);
+				CA_FOREACH_END()
+				if (canPickup && sound != NULL)
+				{
+					e = GameEventNew(GAME_EVENT_SOUND_AT);
+					strcpy(e.u.SoundAt.Sound, sound);
+					e.u.SoundAt.Pos = Vec2ToNet(actor->thing.Pos);
+					GameEventsEnqueue(&gGameEvents, e);
+				}
+			}
+			actor->pickupMenu.pickup = NULL;
+			actor->pickupMenu.effect = NULL;
+			actor->pickupMenu.index = 0;
+			// Reset input
+			KeyLockKeys(&gEventHandlers.keyboard);
+			JoyLock(&gEventHandlers.joysticks);
+			actor->lastCmd = 0;
+			return 0;
+		}
+		else if (Up(cmd) && !Up(actor->lastCmd))
+		{
+			actor->pickupMenu.index--;
+			if (actor->pickupMenu.index < 0)
+			{
+				actor->pickupMenu.index =
+					(int)actor->pickupMenu.effect->u.Menu.Items.size - 1;
+			}
+			SoundPlay(&gSoundDevice, StrSound("menu_switch"));
+		}
+		else if (Down(cmd) && !Down(actor->lastCmd))
+		{
+			actor->pickupMenu.index++;
+			if (actor->pickupMenu.index ==
+				(int)actor->pickupMenu.effect->u.Menu.Items.size)
+			{
+				actor->pickupMenu.index = 0;
+			}
+			SoundPlay(&gSoundDevice, StrSound("menu_switch"));
+		}
 	}
 	else
 	{
-
-		if (actor->confused)
+		actor->hasShot = false;
+		// If this is a pilot, command the vehicle instead
+		if (actor->vehicleUID != -1)
 		{
-			cmd = CmdGetReverse(cmd);
+			TActor *vehicle = ActorGetByUID(actor->vehicleUID);
+			CommandActor(vehicle, cmd, ticks);
+		}
+		else if (actor->pilotUID == -1)
+		{
+			// If this is a vehicle and there's no pilot, do nothing
+		}
+		else
+		{
+
+			if (actor->confused)
+			{
+				cmd = CmdGetReverse(cmd);
+			}
+
+			if (actor->health > 0)
+			{
+				const bool hasChangedDirection =
+					ActorTryChangeDirection(actor, cmd, actor->lastCmd);
+				actor->hasShot = ActorTryShoot(actor, cmd);
+				const bool hasGrenaded = TryGrenade(actor, cmd);
+				const bool hasMoved = ActorTryMove(actor, cmd, ticks);
+				ActorAnimation anim = actor->anim.Type;
+				// Idle if player hasn't done anything
+				if (!(hasChangedDirection || actor->hasShot || hasGrenaded ||
+					  hasMoved))
+				{
+					anim = ACTORANIMATION_IDLE;
+				}
+				else if (hasMoved)
+				{
+					anim = ACTORANIMATION_WALKING;
+				}
+				else
+				{
+					anim = ACTORANIMATION_STAND;
+				}
+				if (actor->anim.Type != anim)
+				{
+					e = GameEventNew(GAME_EVENT_ACTOR_STATE);
+					e.u.ActorState.UID = actor->uid;
+					e.u.ActorState.State = (int32_t)anim;
+					GameEventsEnqueue(&gGameEvents, e);
+				}
+			}
 		}
 
-		if (actor->health > 0)
+		actor->specialCmdDir = CMD_HAS_DIRECTION(cmd);
+		if (Button2(cmd) && !actor->specialCmdDir)
 		{
-			const bool hasChangedDirection =
-				ActorTryChangeDirection(actor, cmd, actor->lastCmd);
-			actor->hasShot = ActorTryShoot(actor, cmd);
-			const bool hasGrenaded = TryGrenade(actor, cmd);
-			const bool hasMoved = ActorTryMove(actor, cmd, ticks);
-			ActorAnimation anim = actor->anim.Type;
-			// Idle if player hasn't done anything
-			if (!(hasChangedDirection || actor->hasShot || hasGrenaded ||
-				  hasMoved))
+			// Special: pick up things that can only be picked up on demand
+			if (!actor->PickupAll && !Button2(actor->lastCmd) &&
+				actor->vehicleUID == -1)
 			{
-				anim = ACTORANIMATION_IDLE;
-			}
-			else if (hasMoved)
-			{
-				anim = ACTORANIMATION_WALKING;
-			}
-			else
-			{
-				anim = ACTORANIMATION_STAND;
-			}
-			if (actor->anim.Type != anim)
-			{
-				GameEvent e = GameEventNew(GAME_EVENT_ACTOR_STATE);
-				e.u.ActorState.UID = actor->uid;
-				e.u.ActorState.State = (int32_t)anim;
+				e = GameEventNew(GAME_EVENT_ACTOR_PICKUP_ALL);
+				e.u.ActorPickupAll.UID = actor->uid;
+				e.u.ActorPickupAll.PickupAll = true;
 				GameEventsEnqueue(&gGameEvents, e);
 			}
 		}
 	}
 
-	actor->specialCmdDir = CMD_HAS_DIRECTION(cmd);
-	if ((cmd & CMD_BUTTON2) && !actor->specialCmdDir)
-	{
-		// Special: pick up things that can only be picked up on demand
-		if (!actor->PickupAll && !(actor->lastCmd & CMD_BUTTON2) &&
-			actor->vehicleUID == -1)
-		{
-			GameEvent e = GameEventNew(GAME_EVENT_ACTOR_PICKUP_ALL);
-			e.u.ActorPickupAll.UID = actor->uid;
-			e.u.ActorPickupAll.PickupAll = true;
-			GameEventsEnqueue(&gGameEvents, e);
-		}
-	}
-
 	actor->lastCmd = cmd;
+	return cmd;
 }
 static bool ActorTryMove(TActor *actor, int cmd, int ticks)
 {
 	const bool canMoveWhenShooting =
-		ConfigGetEnum(&gConfig, "Game.FireMoveStyle") != FIREMOVE_STOP ||
-		!actor->hasShot ||
-		(ConfigGetEnum(&gConfig, "Game.SwitchMoveStyle") ==
-			 SWITCHMOVE_STRAFE &&
-		 (cmd & CMD_BUTTON2));
+		actor->PlayerUID < 0
+			? (actor->flags & FLAGS_MOVE_AND_SHOOT)
+			: (ConfigGetEnum(&gConfig, "Game.FireMoveStyle") !=
+				   FIREMOVE_STOP ||
+			   (ConfigGetEnum(&gConfig, "Game.SwitchMoveStyle") ==
+					SWITCHMOVE_STRAFE &&
+				Button2(cmd)));
+	const bool canMove = !actor->hasShot || canMoveWhenShooting;
 	const bool willMove =
-		!actor->petrified && CMD_HAS_DIRECTION(cmd) && canMoveWhenShooting;
+		!actor->petrified && CMD_HAS_DIRECTION(cmd) && canMove;
 	actor->MoveVel = svec2_zero();
 	if (willMove)
 	{
 		const float moveAmount = ActorGetCharacter(actor)->speed * ticks;
 		struct vec2 moveVel = svec2_zero();
-		if (cmd & CMD_LEFT)
+		if (Left(cmd))
 			moveVel.x--;
-		else if (cmd & CMD_RIGHT)
+		else if (Right(cmd))
 			moveVel.x++;
-		if (cmd & CMD_UP)
+		if (Up(cmd))
 			moveVel.y--;
-		else if (cmd & CMD_DOWN)
+		else if (Down(cmd))
 			moveVel.y++;
 		if (!svec2_is_zero(moveVel))
 		{
@@ -1180,13 +1263,13 @@ void SlideActor(TActor *actor, int cmd)
 	GameEvent e = GameEventNew(GAME_EVENT_ACTOR_SLIDE);
 	e.u.ActorSlide.UID = actor->uid;
 	struct vec2 vel = svec2_zero();
-	if (cmd & CMD_LEFT)
+	if (Left(cmd))
 		vel.x = -SLIDE_X;
-	else if (cmd & CMD_RIGHT)
+	else if (Right(cmd))
 		vel.x = SLIDE_X;
-	if (cmd & CMD_UP)
+	if (Up(cmd))
 		vel.y = -SLIDE_Y;
-	else if (cmd & CMD_DOWN)
+	else if (Down(cmd))
 		vel.y = SLIDE_Y;
 	e.u.ActorSlide.Vel = Vec2ToNet(vel);
 	GameEventsEnqueue(&gGameEvents, e);
@@ -1376,28 +1459,16 @@ static bool CheckManualPickupFunc(
 		strcpy(buttonName, "");
 		InputGetButtonName(
 			pData->inputDevice, pData->deviceIndex, CMD_BUTTON2, buttonName);
-		// TODO: PickupGetName
-		const char *pickupName = NULL;
-		CA_FOREACH(const PickupEffect, pe, p->class->Effects)
-		switch (pe->Type)
-		{
-		case PICKUP_AMMO:
-			pickupName = AmmoGetById(&gAmmo, pe->u.Ammo.Id)->Name;
-			break;
-		case PICKUP_GUN:
-			pickupName = IdWeaponClass(pe->u.GunId)->name;
-			break;
-		default:
-			break;
-		}
-		if (pickupName != NULL)
-		{
-			break;
-		}
-		CA_FOREACH_END()
-		CASSERT(pickupName != NULL, "unknown pickup name");
+		const char *pickupName = PickupClassGetName(p->class);
 		char buf[256];
 		sprintf(buf, "%s to pick up\n%s", buttonName, pickupName);
+		// Hack: remove anything beyond a "+" symbol, so we can have pickups
+		// with the same name but with an invisible suffix
+		char *plus = strrchr(buf, '+');
+		if (plus)
+		{
+			*plus = '\0';
+		}
 		ActorSetChatter(a, buf, 2);
 	}
 	// If co-op AI, alert it so it can try to pick the gun up
@@ -1470,6 +1541,9 @@ static void ActorDie(TActor *actor)
 	GameEvent e = GameEventNew(GAME_EVENT_ACTOR_DIE);
 	e.u.ActorDie.UID = actor->uid;
 	GameEventsEnqueue(&gGameEvents, e);
+
+	// Persist player ammo for when they respawn
+	ActorPersistPlayerWeaponsAndAmmo(actor);
 }
 static bool IsUnarmedBot(const TActor *actor);
 static void ActorAddAmmoPickup(const TActor *actor)
@@ -1683,16 +1757,16 @@ TActor *ActorAdd(NActorAdd aa)
 			AmmoGetById(&gAmmo, i)->Amount * AMMO_STARTING_MULTIPLE;
 		CArrayPushBack(&actor->ammo, &amount);
 	}
-	if (gMission.missionData->WeaponPersist)
+	if (gMission.missionData && gMission.missionData->WeaponPersist)
 	{
 		for (int i = 0; i < aa.Ammo_count; i++)
 		{
 			// Use persisted ammo amount if it is greater,
 			// or if buy/sell is enabled
-			if (gCampaign.Setting.BuyAndSell ||
-				(int)aa.Ammo[i].Amount >
-					AmmoGetById(&gAmmo, aa.Ammo[i].Id)->Amount *
-						AMMO_STARTING_MULTIPLE)
+			const Ammo *ammo = AmmoGetById(&gAmmo, aa.Ammo[i].Id);
+			if (ammo && (gCampaign.Setting.BuyAndSell ||
+						 (int)aa.Ammo[i].Amount >
+							 ammo->Amount * AMMO_STARTING_MULTIPLE))
 			{
 				CArraySet(&actor->ammo, aa.Ammo[i].Id, &aa.Ammo[i].Amount);
 			}
@@ -1722,6 +1796,11 @@ TActor *ActorAdd(NActorAdd aa)
 		Weapon gun = WeaponCreate(c->Gun);
 		actor->guns[0] = gun;
 		actor->gunIndex = 0;
+		if (c->Melee)
+		{
+			Weapon melee = WeaponCreate(c->Melee);
+			actor->guns[MELEE_SLOT] = melee;
+		}
 	}
 	actor->health = aa.Health;
 	actor->action = ACTORACTION_MOVING;
@@ -1770,12 +1849,22 @@ TActor *ActorAdd(NActorAdd aa)
 		ActorSetAIState(actor, AI_STATE_IDLE);
 	}
 
-	EmitterInit(
-		&actor->barrelSmoke, StrParticleClass(&gParticleClasses, "smoke"),
-		svec2_zero(), -0.05f, 0.05f, 3, 3, 0, 0, 10);
-	EmitterInit(
-		&actor->healEffect, StrParticleClass(&gParticleClasses, "health_plus"),
-		svec2_zero(), -0.1f, 0.1f, 0, 0, 0, 0, 0);
+	const ParticleClass *barrelSmoke =
+		StrParticleClass(&gParticleClasses, "smoke");
+	if (barrelSmoke)
+	{
+		EmitterInit(
+			&actor->barrelSmoke, barrelSmoke, svec2_zero(), -0.05f, 0.05f, 3,
+			3, 0, 0, 10);
+	}
+	const ParticleClass *healthPlus =
+		StrParticleClass(&gParticleClasses, "health_plus");
+	if (healthPlus)
+	{
+		EmitterInit(
+			&actor->healEffect, healthPlus, svec2_zero(), -0.1f, 0.1f, 0, 0, 0,
+			0, 0);
+	}
 	GoreEmitterInit(&actor->blood1, "blood1");
 	GoreEmitterInit(&actor->blood2, "blood2");
 	GoreEmitterInit(&actor->blood3, "blood3");
@@ -1845,6 +1934,10 @@ struct vec2 ActorGetMuzzleOffset(
 int ActorWeaponGetAmmo(
 	const TActor *a, const WeaponClass *wc, const int barrel)
 {
+	if (wc == NULL)
+	{
+		return 0;
+	}
 	const int ammoId = WC_BARREL_ATTR(*wc, AmmoId, barrel);
 	if (ammoId == -1)
 	{
@@ -1871,7 +1964,7 @@ int ActorGetCanFireBarrel(const TActor *a, const Weapon *w)
 	}
 	return barrel;
 }
-static bool CanSwitchToWeapon(const TActor *a, const int slot)
+bool ActorCanSwitchToWeapon(const TActor *a, const int slot)
 {
 	const WeaponClass *wc = a->guns[slot].Gun;
 	if (wc == NULL)
@@ -1901,20 +1994,23 @@ static bool CanSwitchToWeapon(const TActor *a, const int slot)
 			}
 		}
 	}
+	// Don't switch to grenade, unless the player has no grenade button
+	if (slot == GRENADE_SLOT)
+	{
+		const PlayerData *p = PlayerDataGetByUID(a->PlayerUID);
+		return !PlayerHasGrenadeButton(p);
+	}
 	return true;
 }
-bool ActorTrySwitchWeapon(const TActor *a, const bool allGuns)
+bool ActorTrySwitchWeapon(const TActor *a)
 {
 	// Find the next weapon to switch to
-	// If the player does not have a grenade key set, allow switching to
-	// grenades (classic style)
-	const int switchCount = allGuns ? MAX_WEAPONS : MAX_GUNS;
 	const int startIndex = ActorGetNumGuns(a) > 0 ? a->gunIndex : MAX_GUNS;
 	int weaponIndex = startIndex;
 	do
 	{
-		weaponIndex = (weaponIndex + 1) % switchCount;
-	} while (!CanSwitchToWeapon(a, weaponIndex));
+		weaponIndex = (weaponIndex + 1) % MAX_WEAPONS;
+	} while (!ActorCanSwitchToWeapon(a, weaponIndex));
 	if (weaponIndex == startIndex)
 	{
 		// No other weapon to switch to
@@ -1939,7 +2035,86 @@ void ActorSwitchGun(const NActorSwitchGun sg)
 	}
 	const WeaponClass *gun = ACTOR_GET_WEAPON(a)->Gun;
 	SoundPlayAt(&gSoundDevice, gun->SwitchSound, a->thing.Pos);
-	ActorSetChatter(a, gun->name, CHATTER_SWITCH_GUN);
+	// Show gun switch text as chatter, but only for PCs
+	if (a->PlayerUID >= 0)
+	{
+		ActorSetChatter(a, gun->name, CHATTER_SWITCH_GUN);
+	}
+}
+
+void ActorPickupGun(const TActor *a, const WeaponClass *wc)
+{
+	int actorsGunIdx = ActorFindGun(a, wc);
+	if (actorsGunIdx >= 0)
+	{
+		// Actor already has gun
+
+		// Switch to the same gun
+		if (ActorCanSwitchToWeapon(a, actorsGunIdx))
+		{
+			GameEvent e = GameEventNew(GAME_EVENT_ACTOR_SWITCH_GUN);
+			e.u.ActorSwitchGun.UID = a->uid;
+			e.u.ActorSwitchGun.GunIdx = actorsGunIdx;
+			GameEventsEnqueue(&gGameEvents, e);
+		}
+
+		// Drop the same gun
+		PickupAddGun(wc, a->Pos);
+	}
+	else
+	{
+		// Make sure weapons are picked up in the correct slot
+		int slotStart, slotEnd;
+		GunTypeGetSlotStartEnd(wc->Type, &slotStart, &slotEnd);
+		if (actorsGunIdx < slotStart || actorsGunIdx > slotEnd)
+		{
+			actorsGunIdx = slotStart;
+		}
+
+		// Pickup gun
+		// Replace the current gun, unless there's a free slot, in which case
+		// pick up into the free spot
+		GameEvent e = GameEventNew(GAME_EVENT_ACTOR_REPLACE_GUN);
+		e.u.ActorReplaceGun.UID = a->uid;
+		strcpy(e.u.ActorReplaceGun.Gun, wc->name);
+		switch (wc->Type)
+		{
+		case GUNTYPE_MELEE:
+			e.u.ActorReplaceGun.GunIdx = MELEE_SLOT;
+			break;
+		case GUNTYPE_GRENADE:
+			e.u.ActorReplaceGun.GunIdx = GRENADE_SLOT;
+			break;
+		default:
+			if (a->gunIndex < MELEE_SLOT)
+			{
+				e.u.ActorReplaceGun.GunIdx = a->gunIndex;
+			}
+			else
+			{
+				e.u.ActorReplaceGun.GunIdx = 0;
+			}
+			break;
+		}
+		int replaceGunIndex = e.u.ActorReplaceGun.GunIdx;
+		for (int i = slotStart; i <= slotEnd; i++)
+		{
+			if (a->guns[i].Gun == NULL)
+			{
+				e.u.ActorReplaceGun.GunIdx = i;
+				replaceGunIndex = -1;
+				break;
+			}
+		}
+		GameEventsEnqueue(&gGameEvents, e);
+
+		// If replacing a gun, "drop" the gun being replaced (i.e. create a gun
+		// pickup)
+		if (replaceGunIndex >= 0)
+		{
+			PickupAddGun(a->guns[replaceGunIndex].Gun, a->Pos);
+		}
+	}
 }
 
 bool ActorIsImmune(const TActor *actor, const special_damage_e damage)
@@ -2007,16 +2182,14 @@ static void ActorTakeSpecialDamage(
 	}
 }
 
-static void ActorTakeHit(
-	TActor *actor, const int flags, const int sourceUID,
-	const special_damage_e damage, const int specialTicks);
+static void ActorTakeHit(TActor *actor, const NThingDamage d);
 void ActorHit(const NThingDamage d)
 {
 	TActor *a = ActorGetByUID(d.UID);
 	if (!a->isInUse)
 		return;
 
-	ActorTakeHit(a, d.Flags, d.SourceActorUID, d.Special, d.SpecialTicks);
+	ActorTakeHit(a, d);
 	if (d.Power > 0)
 	{
 		DamageActor(a, d.Power, d.SourceActorUID);
@@ -2043,6 +2216,7 @@ void ActorHit(const NThingDamage d)
 		}
 		CA_FOREACH_END()
 		a->accumulatedDamage = damage;
+		a->damageCooldownTicks = 0;
 
 		GameEvent s = GameEventNew(GAME_EVENT_ADD_PARTICLE);
 		s.u.AddParticle.Class =
@@ -2070,23 +2244,31 @@ void ActorHit(const NThingDamage d)
 	}
 }
 
-static void ActorTakeHit(
-	TActor *actor, const int flags, const int sourceUID,
-	const special_damage_e damage, const int specialTicks)
+static void ActorTakeHit(TActor *actor, const NThingDamage d)
 {
 	// Wake up if this is an AI
 	if (!gCampaign.IsClient)
 	{
 		AIWake(actor, 1);
 	}
-	const TActor *source = ActorGetByUID(sourceUID);
+	const TActor *source = ActorGetByUID(d.SourceActorUID);
 	const int playerUID = source != NULL ? source->PlayerUID : -1;
 	if (ActorIsInvulnerable(
-			actor, flags, playerUID, gCampaign.Entry.Mode, damage))
+			actor, d.Flags, playerUID, gCampaign.Entry.Mode, d.Special))
 	{
 		return;
 	}
-	ActorTakeSpecialDamage(actor, damage, specialTicks);
+	ActorTakeSpecialDamage(actor, d.Special, d.SpecialTicks);
+
+	// Update hits
+	// TODO: correct accuracy for persistent bullets
+	if (!gCampaign.IsClient && playerUID >= 0 &&
+		strlen(d.SourceWeaponClassName) > 0)
+	{
+		PlayerData *p = PlayerDataGetByUID(playerUID);
+		const WeaponClass *wc = StrWeaponClass(d.SourceWeaponClassName);
+		WeaponUsagesUpdate(p->WeaponUsages, wc, 0, 1);
+	}
 }
 
 bool ActorIsInvulnerable(
@@ -2134,6 +2316,11 @@ static void ActorAddBloodSplatters(
 	const GoreAmount ga = ConfigGetEnum(&gConfig, "Graphics.Gore");
 	if (ga == GORE_NONE)
 		return;
+	const color_t bloodColor = ActorGetCharacter(a)->Class->BloodColor;
+	if (bloodColor.a == 0)
+	{
+		return;
+	}
 
 	// Emit blood based on power and gore setting
 	int bloodPower = power * 2;
@@ -2169,7 +2356,7 @@ static void ActorAddBloodSplatters(
 		ap.Angle = NAN;
 		ap.Z = 10;
 		ap.Vel = vel;
-		ap.Mask = ActorGetCharacter(a)->Class->BloodColor;
+		ap.Mask = bloodColor;
 		EmitterStart(em, &ap);
 		switch (ga)
 		{
@@ -2186,6 +2373,12 @@ static void ActorAddBloodSplatters(
 	}
 }
 
+int ActorGetMaxHeal(const TActor *a, const bool exceedMax)
+{
+	const Character *c = ActorGetCharacter(a);
+	return exceedMax ? c->excessHealth : c->maxHealth;
+}
+
 int ActorGetHealthPercent(const TActor *a)
 {
 	const int maxHealth = ActorGetCharacter(a)->maxHealth;
@@ -2226,3 +2419,18 @@ bool ActorIsLocalPlayer(const int uid)
 
 	return PlayerIsLocal(a->PlayerUID);
 }
+
+void ActorPersistPlayerWeaponsAndAmmo(const TActor *a)
+{
+	// Update the player's weapons and ammo based on their actor
+	PlayerData *p = PlayerDataGetByUID(a->PlayerUID);
+	if (p == NULL)
+	{
+		return;
+	}
+	for (int i = 0; i < MAX_WEAPONS; i++)
+	{
+		p->guns[i] = a->guns[i].Gun;
+	}
+	CArrayCopy(&p->ammo, &a->ammo);
+}
