@@ -22,7 +22,7 @@
 	This file incorporates work covered by the following copyright and
 	permission notice:
 
-	Copyright (c) 2013-2021 Cong Xu
+	Copyright (c) 2013-2021, 2024-2025 Cong Xu
 	All rights reserved.
 
 	Redistribution and use in source and binary forms, with or without
@@ -133,6 +133,18 @@ void DamageObject(const NThingDamage d)
 			EmitterStart(&em, &ap);
 		}
 	}
+
+	// Update hits
+	// TODO: correct accuracy for persistent bullets
+	const TActor *source = ActorGetByUID(d.SourceActorUID);
+	const int playerUID = source != NULL ? source->PlayerUID : -1;
+	if (!gCampaign.IsClient && playerUID >= 0 &&
+		strlen(d.SourceWeaponClassName) > 0)
+	{
+		PlayerData *p = PlayerDataGetByUID(playerUID);
+		const WeaponClass *wc = StrWeaponClass(d.SourceWeaponClassName);
+		WeaponUsagesUpdate(p->WeaponUsages, wc, 0, 1);
+	}
 }
 
 static void AddPickupAtObject(const TObject *o, const PickupType type)
@@ -221,7 +233,6 @@ void ObjRemove(const NMapObjectRemove mor)
 		if (strlen(o->Class->Wreck.Bullet) > 0)
 		{
 			// A wreck left after the destruction of this object
-			// TODO: doesn't need to be network event
 			GameEvent e = GameEventNew(GAME_EVENT_ADD_BULLET);
 			e.u.AddBullet.UID = MobObjsObjsGetNextUID();
 			strcpy(e.u.AddBullet.BulletClass, o->Class->Wreck.Bullet);
@@ -302,25 +313,27 @@ bool HasHitSound(
 
 static void DoDamageThing(
 	const ThingKind targetKind, const int targetUID, const TActor *source,
-	const int flags, const BulletClass *bullet, const bool canDamage,
-	const struct vec2 hitVector);
+	const int flags, const BulletClass *bullet, const WeaponClass *weapon,
+	const bool canDamage, const struct vec2 hitVector);
 static void DoDamageCharacter(
 	const TActor *actor, const TActor *source, const struct vec2 hitVector,
-	const BulletClass *bullet, const int flags);
+	const BulletClass *bullet, const WeaponClass *weapon, const int flags);
 void Damage(
-	const struct vec2 hitVector, const BulletClass *bullet, const int flags,
-	const TActor *source, const ThingKind targetKind, const int targetUID)
+	const struct vec2 hitVector, const BulletClass *bullet,
+	const WeaponClass *weapon, const int flags, const TActor *source,
+	const ThingKind targetKind, const int targetUID)
 {
 	switch (targetKind)
 	{
 	case KIND_CHARACTER: {
 		const TActor *actor = ActorGetByUID(targetUID);
-		DoDamageCharacter(actor, source, hitVector, bullet, flags);
+		DoDamageCharacter(actor, source, hitVector, bullet, weapon, flags);
 	}
 	break;
 	case KIND_OBJECT:
 		DoDamageThing(
-			targetKind, targetUID, source, flags, bullet, true, hitVector);
+			targetKind, targetUID, source, flags, bullet, weapon, true,
+			hitVector);
 		break;
 	default:
 		CASSERT(false, "cannot damage tile item kind");
@@ -329,8 +342,8 @@ void Damage(
 }
 static void DoDamageThing(
 	const ThingKind targetKind, const int targetUID, const TActor *source,
-	const int flags, const BulletClass *bullet, const bool canDamage,
-	const struct vec2 hitVector)
+	const int flags, const BulletClass *bullet, const WeaponClass *weapon,
+	const bool canDamage, const struct vec2 hitVector)
 {
 	GameEvent e = GameEventNew(GAME_EVENT_THING_DAMAGE);
 	e.u.ThingDamage.UID = targetUID;
@@ -343,11 +356,15 @@ static void DoDamageThing(
 		e.u.ThingDamage.Power = 0;
 	}
 	e.u.ThingDamage.Vel = Vec2ToNet(hitVector);
+	if (weapon)
+	{
+		strcpy(e.u.ThingDamage.SourceWeaponClassName, weapon->name);
+	}
 	GameEventsEnqueue(&gGameEvents, e);
 }
 static void DoDamageCharacter(
 	const TActor *actor, const TActor *source, const struct vec2 hitVector,
-	const BulletClass *bullet, const int flags)
+	const BulletClass *bullet, const WeaponClass *weapon, const int flags)
 {
 	// Create events: hit, damage, score
 	CASSERT(actor->isInUse, "Cannot damage nonexistent player");
@@ -373,15 +390,14 @@ static void DoDamageCharacter(
 		CanDamageCharacter(flags, source, actor, bullet->Special.Effect);
 
 	DoDamageThing(
-		KIND_CHARACTER, actor->uid, source, flags, bullet, canDamage,
+		KIND_CHARACTER, actor->uid, source, flags, bullet, weapon, canDamage,
 		hitVector);
 
 	if (canDamage)
 	{
 		// Don't score for friendly, unpiloted vehicle, or player hits
 		const bool isFriendly =
-			(actor->flags & FLAGS_GOOD_GUY) ||
-			actor->pilotUID == -1 ||
+			(actor->flags & FLAGS_GOOD_GUY) || actor->pilotUID == -1 ||
 			(!IsPVP(gCampaign.Entry.Mode) && actor->PlayerUID >= 0);
 		if (source && source->PlayerUID >= 0 && bullet->Power != 0 &&
 			!isFriendly)
@@ -499,7 +515,9 @@ void ObjAdd(const NMapObjectAdd amo)
 	o->thing.CPicFunc = MapObjectDraw;
 	MapTryMoveThing(&gMap, &o->thing, NetToVec2(amo.Pos));
 	EmitterInit(
-		&o->damageSmoke, StrParticleClass(&gParticleClasses, "smoke_big"),
+		&o->damageSmoke,
+		StrParticleClass(
+			&gParticleClasses, o->Class->DamageSmoke.ParticleClass),
 		svec2_zero(), -0.05f, 0.05f, 3, 3, 0, 0, 20);
 	o->isInUse = true;
 	LOG(LM_MAIN, LL_DEBUG,
@@ -550,6 +568,7 @@ void UpdateObjects(const int ticks)
 					RAND_FLOAT(-obj->thing.size.x / 4, obj->thing.size.x / 4),
 					RAND_FLOAT(
 						-obj->thing.size.y / 4, obj->thing.size.y / 4)));
+			ap.Z = obj->Class->DamageSmoke.Z;
 			ap.Mask = colorWhite;
 			EmitterUpdate(&obj->damageSmoke, &ap, ticks);
 		}
@@ -593,7 +612,7 @@ void UpdateObjects(const int ticks)
 				CArrayGet(
 					&gCampaign.Setting.characters.OtherChars,
 					obj->Class->u.Character.CharId),
-				true);
+				NULL);
 			e.u.ActorAdd.CharId = obj->Class->u.Character.CharId;
 			GameEventsEnqueue(&gGameEvents, e);
 
