@@ -1,7 +1,7 @@
 /*
 	C-Dogs SDL
 	A port of the legendary (and fun) action/arcade cdogs.
-	Copyright (c) 2013-2020 Cong Xu
+	Copyright (c) 2013-2020, 2024 Cong Xu
 	All rights reserved.
 
 	Redistribution and use in source and binary forms, with or without
@@ -39,6 +39,7 @@
 #include "log.h"
 
 #define GRAPHICS_DIR "graphics"
+#define GRAPHICS_HD_DIR "graphics_hd"
 
 PicManager gPicManager;
 
@@ -54,7 +55,10 @@ void PicManagerInit(PicManager *pm)
 	pm->sprites = hashmap_new();
 	pm->customPics = hashmap_new();
 	pm->customSprites = hashmap_new();
-	CArrayInit(&pm->hairstyleNames, sizeof(char *));
+	for (HeadPart hp = HEAD_PART_HAIR; hp < HEAD_PART_COUNT; hp++)
+	{
+		CArrayInit(&pm->headPartNames[hp], sizeof(char *));
+	}
 	CArrayInit(&pm->wallStyleNames, sizeof(char *));
 	CArrayInit(&pm->tileStyleNames, sizeof(char *));
 	CArrayInit(&pm->exitStyleNames, sizeof(char *));
@@ -66,7 +70,8 @@ static NamedPic *AddNamedPic(map_t pics, const char *name, const Pic *p);
 static NamedSprites *AddNamedSprites(map_t sprites, const char *name);
 static void AfterAdd(PicManager *pm);
 static void PicManagerAdd(
-	map_t pics, map_t sprites, const char *name, SDL_Surface *imageIn)
+	map_t pics, map_t sprites, const char *name, SDL_Surface *imageIn,
+	const bool isHD)
 {
 	char buf[CDOGS_FILENAME_MAX];
 	const char *dot = strrchr(name, '.');
@@ -131,10 +136,28 @@ static void PicManagerAdd(
 			{
 				pic = &np->pic;
 			}
-			PicLoad(pic, size, offset, image);
+			PicLoad(pic, size, offset, image, isHD);
 
 			if (strncmp("chars/", buf, strlen("chars/")) == 0)
 			{
+				// All head parts use hair color, so determine
+				// which head part we are looking at
+				const char *subfolder = buf + strlen("chars/");
+				CharColorType headPartColor = CHAR_COLOR_HAIR;
+				if (strncmp("facehairs/", subfolder, strlen("facehairs/")) ==
+					0)
+				{
+					headPartColor = CHAR_COLOR_FACEHAIR;
+				}
+				else if (strncmp("hats/", subfolder, strlen("hats/")) == 0)
+				{
+					headPartColor = CHAR_COLOR_HAT;
+				}
+				else if (
+					strncmp("glasses/", subfolder, strlen("glasses/")) == 0)
+				{
+					headPartColor = CHAR_COLOR_GLASSES;
+				}
 				// Convert char pics to multichannel version
 				for (int i = 0; i < pic->size.x * pic->size.y; i++)
 				{
@@ -147,7 +170,8 @@ static void PicManagerAdd(
 					}
 					// Convert character color keyed color to
 					// greyscale + special alpha
-					const CharColorType colorType = CharColorTypeFromColor(c);
+					const CharColorType colorType =
+						CharColorTypeFromColor(c, headPartColor);
 					color_t converted = c;
 					if (colorType != CHAR_COLOR_COUNT)
 					{
@@ -168,7 +192,7 @@ static void PicManagerAdd(
 
 void PicManagerLoadDir(
 	PicManager *pm, const char *path, const char *prefix, map_t pics,
-	map_t sprites)
+	map_t sprites, const bool isHD)
 {
 	tinydir_dir dir;
 	if (tinydir_open(&dir, path) == -1)
@@ -215,7 +239,7 @@ void PicManagerLoadDir(
 					{
 						PathGetBasenameWithoutExtension(buf, file.name);
 					}
-					PicManagerAdd(pics, sprites, buf, data);
+					PicManagerAdd(pics, sprites, buf, data, isHD);
 				}
 			}
 			rwops->close(rwops);
@@ -226,11 +250,12 @@ void PicManagerLoadDir(
 			{
 				char buf[CDOGS_PATH_MAX];
 				sprintf(buf, "%s/%s", prefix, file.name);
-				PicManagerLoadDir(pm, file.path, buf, pics, sprites);
+				PicManagerLoadDir(pm, file.path, buf, pics, sprites, isHD);
 			}
 			else
 			{
-				PicManagerLoadDir(pm, file.path, file.name, pics, sprites);
+				PicManagerLoadDir(
+					pm, file.path, file.name, pics, sprites, isHD);
 			}
 		}
 	}
@@ -242,14 +267,19 @@ void PicManagerLoad(PicManager *pm)
 {
 	char buf[CDOGS_PATH_MAX];
 	GetDataFilePath(buf, GRAPHICS_DIR);
-	PicManagerLoadDir(pm, buf, NULL, pm->pics, pm->sprites);
+	PicManagerLoadDir(pm, buf, NULL, pm->pics, pm->sprites, false);
+	GetDataFilePath(buf, GRAPHICS_HD_DIR);
+	PicManagerLoadDir(pm, buf, NULL, pm->pics, pm->sprites, true);
 }
 
 static void FindStylePics(
 	PicManager *pm, CArray *styleNames, PFany hashmapFunc);
 static void FindStyleSprites(
 	PicManager *pm, CArray *styleNames, PFany hashmapFunc);
-static int MaybeAdHairSpriteName(any_t data, any_t item);
+static int MaybeAddHairSpriteName(any_t data, any_t item);
+static int MaybeAddFacehairSpriteName(any_t data, any_t item);
+static int MaybeAddHatSpriteName(any_t data, any_t item);
+static int MaybeAddGlassesSpriteName(any_t data, any_t item);
 static int MaybeAddWallPicName(any_t data, any_t item);
 static int MaybeAddTilePicName(any_t data, any_t item);
 static int MaybeAddExitPicName(any_t data, any_t item);
@@ -257,7 +287,15 @@ static int MaybeAddKeyPicName(any_t data, any_t item);
 static int MaybeAddDoorPicName(any_t data, any_t item);
 static void AfterAdd(PicManager *pm)
 {
-	FindStyleSprites(pm, &pm->hairstyleNames, MaybeAdHairSpriteName);
+	FindStyleSprites(
+		pm, &pm->headPartNames[HEAD_PART_HAIR], MaybeAddHairSpriteName);
+	FindStyleSprites(
+		pm, &pm->headPartNames[HEAD_PART_FACEHAIR],
+		MaybeAddFacehairSpriteName);
+	FindStyleSprites(
+		pm, &pm->headPartNames[HEAD_PART_HAT], MaybeAddHatSpriteName);
+	FindStyleSprites(
+		pm, &pm->headPartNames[HEAD_PART_GLASSES], MaybeAddGlassesSpriteName);
 	FindStylePics(pm, &pm->wallStyleNames, MaybeAddWallPicName);
 	FindStylePics(pm, &pm->tileStyleNames, MaybeAddTilePicName);
 	FindStylePics(pm, &pm->exitStyleNames, MaybeAddExitPicName);
@@ -344,14 +382,34 @@ static void MaybeAddStyleName(
 	CSTRDUP(s, buf);
 	CArrayPushBack(styleNames, &s);
 }
-static int MaybeAdHairSpriteName(any_t data, any_t item)
+static int MaybeAddHeadPartSpriteName(
+	any_t data, any_t item, const HeadPart hp, const char *path)
 {
 	PicManager *pm = data;
 	MaybeAddStyleName(
-		((const NamedSprites *)item)->name, "chars/hairs/",
-		&pm->hairstyleNames);
+		((const NamedSprites *)item)->name, path, &pm->headPartNames[hp]);
 	return MAP_OK;
 }
+static int MaybeAddHairSpriteName(any_t data, any_t item)
+{
+	return MaybeAddHeadPartSpriteName(
+		data, item, HEAD_PART_HAIR, "chars/hairs/");
+}
+static int MaybeAddFacehairSpriteName(any_t data, any_t item)
+{
+	return MaybeAddHeadPartSpriteName(
+		data, item, HEAD_PART_FACEHAIR, "chars/facehairs/");
+}
+static int MaybeAddHatSpriteName(any_t data, any_t item)
+{
+	return MaybeAddHeadPartSpriteName(
+		data, item, HEAD_PART_HAT, "chars/hats/");
+}
+static int MaybeAddGlassesSpriteName(any_t data, any_t item)
+{
+	return MaybeAddHeadPartSpriteName(
+		data, item, HEAD_PART_GLASSES, "chars/glasses/");
+}
 static int MaybeAddExitPicName(any_t data, any_t item)
 {
 	// Exit pics should be like:
@@ -434,7 +492,10 @@ static void StyleNamesDestroy(CArray *a)
 void PicManagerTerminate(PicManager *pm)
 {
 	PicManagerUnload(pm);
-	StyleNamesDestroy(&pm->hairstyleNames);
+	for (HeadPart hp = HEAD_PART_HAIR; hp < HEAD_PART_COUNT; hp++)
+	{
+		StyleNamesDestroy(&pm->headPartNames[hp]);
+	}
 	StyleNamesDestroy(&pm->wallStyleNames);
 	StyleNamesDestroy(&pm->tileStyleNames);
 	StyleNamesDestroy(&pm->exitStyleNames);
@@ -458,10 +519,10 @@ static int ReloadTexture(any_t data, any_t item);
 static int ReloadSpriteTexture(any_t data, any_t item);
 void PicManagerReloadTextures(PicManager *pm)
 {
-	hashmap_iterate(pm->pics, ReloadTexture, pm);
-	hashmap_iterate(pm->customPics, ReloadTexture, pm);
-	hashmap_iterate(pm->sprites, ReloadSpriteTexture, pm);
-	hashmap_iterate(pm->customSprites, ReloadSpriteTexture, pm);
+	hashmap_iterate(pm->pics, ReloadTexture, NULL);
+	hashmap_iterate(pm->customPics, ReloadTexture, NULL);
+	hashmap_iterate(pm->sprites, ReloadSpriteTexture, NULL);
+	hashmap_iterate(pm->customSprites, ReloadSpriteTexture, NULL);
 }
 static int ReloadTexture(any_t data, any_t item)
 {
