@@ -22,7 +22,7 @@
 	This file incorporates work covered by the following copyright and
 	permission notice:
 
-	Copyright (c) 2013-2017, 2019-2022 Cong Xu
+	Copyright (c) 2013-2017, 2019-2022, 2024 Cong Xu
 	All rights reserved.
 
 	Redistribution and use in source and binary forms, with or without
@@ -62,6 +62,7 @@
 
 #include "events.h"
 #include "joystick.h"
+#include "log.h"
 #include "sys_config.h"
 
 bool gTrue = true;
@@ -234,12 +235,6 @@ void RealPath(const char *src, char *dest)
 		{
 			*c = '/';
 		}
-#ifdef _WIN32
-		else
-		{
-			*c = (char)toupper(*c);
-		}
-#endif
 	}
 	TrimSlashes(dest);
 }
@@ -606,10 +601,26 @@ bool IntsEqual(const void *v1, const void *v2)
 	return *(const int *)v1 == *(const int *)v2;
 }
 
+const char *HeadPartStr(const HeadPart hp)
+{
+	switch (hp)
+	{
+		T2S(HEAD_PART_HAIR, "Hair");
+		T2S(HEAD_PART_FACEHAIR, "Facial Hair");
+		T2S(HEAD_PART_HAT, "Hat");
+		T2S(HEAD_PART_GLASSES, "Glasses");
+	default:
+		return "";
+	}
+}
+
 BodyPart StrBodyPart(const char *s)
 {
 	S2T(BODY_PART_HEAD, "head");
 	S2T(BODY_PART_HAIR, "hair");
+	S2T(BODY_PART_FACEHAIR, "facehair");
+	S2T(BODY_PART_HAT, "hat");
+	S2T(BODY_PART_GLASSES, "glasses");
 	S2T(BODY_PART_BODY, "body");
 	S2T(BODY_PART_LEGS, "legs");
 	S2T(BODY_PART_GUN_R, "gun_r");
@@ -627,3 +638,47 @@ int Pulse256(const int t)
 	}
 	return alphaUnscaled;
 }
+
+char *ReadFileIntoBuf(const char *path, const char *mode)
+{
+	char *buf = NULL;
+	FILE *f = fopen(path, mode);
+	if (f == NULL)
+	{
+		goto bail;
+	}
+
+	// Read into buffer
+	if (fseek(f, 0L, SEEK_END) != 0)
+	{
+		goto bail;
+	}
+	long len = ftell(f);
+	if (len == -1)
+	{
+		goto bail;
+	}
+	CCALLOC(buf, len + 1);
+	if (fseek(f, 0L, SEEK_SET) != 0)
+	{
+		goto bail;
+	}
+	if (fread(buf, 1, len, f) == 0)
+	{
+		goto bail;
+	}
+
+	goto end;
+
+bail:
+	CFREE(buf);
+	buf = NULL;
+
+end:
+	if (f != NULL && fclose(f) != 0)
+	{
+		LOG(LM_MAIN, LL_ERROR, "Cannot close file %s: %s", path,
+			strerror(errno));
+	}
+	return buf;
+}
